"""Cache of Saku BBS.
"""
#
# Copyright (c) 2005,2006 shinGETsu Project.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
# OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
# OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
# SUCH DAMAGE.
#
# $Id: cache.py 1281 2006-09-15 23:23:47Z fuktommy $
#

import base64
import md5
import os
import random
import re
import shutil
import sys
from time import time

if hasattr(sys, "winver"):
    from os import listdir
else:
    from dircache import listdir

import apollo
import config
import spam
from node import *

__version__ = "$Revision: 1281 $"
__all__ = ["Record", "Cache", "CacheList", "UpdateList", "RecentList"]

class Record(dict):

    """One article of BBS.

    timestamp<>id<>foo:bar<>hoge:fuga<>...
    """

    recstr = ""     # record in string format
    datfile = ""    # cache file
    stamp = 0       # date when the record made
    id = ""         # md5sum
    idstr = ""      # stamp_id
    path = ""       # path for real file
    body_path = ''  # path for body (record without attach field)
    rm_path = ""    # path for removed marker
    flag_load = False
    flag_load_body = False

    def __init__(self, datfile="", idstr=""):
        dict.__init__(self)
        self.datfile = datfile
        self.idstr = idstr

        if idstr != "":
            buf = idstr.split("_")
            try:
                self.stamp = int(buf[0])
                self.id = buf[1]
            except ValueError:
                sys.stderr.write('%s: bad format\n' % idstr)
                pass
        self.setpath()

    def free(self):
        self.flag_load = False
        self.flag_load_body = False
        for key in self.keys():
            del self[key]

    def __str__(self):
        return self.recstr

    def __gt__(self, y):
        if self.stamp != y.stamp:
            return self.stamp > y.stamp
        else:
            return self.idstr > y.idstr

    def __lt__(self, y):
        if self.stamp != y.stamp:
            return self.stamp < y.stamp
        else:
            return self.idstr < y.idstr

    def __eq__(self, y):
        return self.idstr == y.idstr

    def __ne__(self, y):
        return self.idstr != y.idstr

    def setpath(self):
        if (self.idstr != "") and (self.datfile != ""):
            self.path = os.path.join(config.cache_dir,
                                     self.datfile,
                                     'record',
                                     self.idstr)
            self.body_path = os.path.join(config.cache_dir,
                                          self.datfile,
                                          'body',
                                          self.idstr)
            self.rm_path = os.path.join(config.cache_dir,
                                        self.datfile,
                                        'removed',
                                        self.idstr)

    def parse(self, recstr):
        """Parse cache record."""
        self.recstr = re.sub(r"[\r\n]*$", "", recstr)
        tmp = self.recstr.split("<>")
        try:
            self["stamp"] = tmp.pop(0)
            self["id"] = tmp.pop(0)
            self.idstr = self["stamp"] + "_" + self["id"]
            self.stamp = int(self["stamp"])
            self.id = self["id"]
        except (ValueError, KeyError, IndexError):
            sys.stderr.write(recstr + ": bad format\n")
            return False
        for i in tmp:
            buf = i.split(":", 1)
            if len(buf) >= 2:
                buf[1] = re.sub(r"<br>", "\n", buf[1])
                buf[1] = re.sub(r"<", "&lt;", buf[1])
                buf[1] = re.sub(r">", "&gt;", buf[1])
                buf[1] = re.sub(r"\n", "<br>", buf[1])
                self[buf[0]] = buf[1]
        self.flag_load = True
        self.flag_load_body = True
        self.setpath()
        return True

    def _load(self, filename):
        if self.path == '':
            sys.stderr.write('Null file name\n')
            return False
        try:
            if self.size() <= 0:
                self.remove()
                return False
            f = file(filename)
            parse_ok = self.parse(f.readline())
            f.close()
            return parse_ok
        except (IOError, OSError), err:
            sys.stderr.write('IOError/OSError: %s\n' % err)
            return False

    def load(self):
        if not self.flag_load:
            return self._load(self.path)
        return True

    def load_body(self):
        if not self.flag_load_body:
            return self._load(self.body_path)
        return True

    def build(self, stamp, body, passwd=""):
        target = body.keys()
        bodyary = []
        for key in target:
            bodyary.append(key + ":" + body[key])
            self[key] = body[key]
        bodystr = "<>".join(bodyary)
        if passwd != "":
            (pubkey, prikey) = apollo.key_pair(passwd)
            md = md5.new(bodystr).hexdigest()
            sign = apollo.sign(md, pubkey, prikey)
            self["pubkey"] = pubkey
            self["sign"] = sign
            self["target"] = ",".join(target)

            bodystr += "<>pubkey:" + pubkey + \
                       "<>sign:" + sign + \
                       "<>target:" + ",".join(target)

        id = md5.new(bodystr).hexdigest()
        self.stamp = int(stamp)
        self.recstr = str(stamp) + "<>" + id + "<>" + bodystr
        self.idstr = str(stamp) + "_" + id
        self["stamp"] = str(stamp)
        self["id"] = id
        self.id = id
        self.setpath()
        return id

    def md5check(self):
        buf = str(self).split("<>", 2)
        if (len(buf) > 2):
            return md5.new(buf[2]).hexdigest() == self["id"]
        else:
            return False

    def attach_path(self, suffix=None):
        if self.path == "":
            sys.stderr.write("Null file name\n")
            return None
        dir = "/".join((config.cache_dir, self.datfile, "attach"))
        if suffix is not None:
            suffix = re.sub(r"[^-_.A-Za-z0-9]", "", suffix)
            if suffix == "":
                suffix = "txt"
            return dir + "/" + self.idstr + "." + suffix
        for i in listdir(dir):
            if i.startswith(self.idstr):
                return dir + "/" + i
        return None

    def attach_size(self, path=None):
        if path is None:
            path = self.attach_path()
        if path is None:
            return 0
        else:
            return os.path.getsize(path)

    def _write_file(self, path, data):
        """Save data to file."""
        if path == "":
            sys.stderr.write("Null file name\n")
            return False
        try:
            f = file(path, "wb")
            f.write(data)
            f.close()
            return True
        except IOError:
            sys.stderr.write(path + ": IOError\n")
            return False

    def sync(self, force=False):
        """Save files."""
        if self.removed():
            return
        elif (not force) and self.exists():
            pass
        else:
            self._write_file(self.path, str(self)+"\n")
        self._write_file(self.body_path, self.body_string()+'\n')
        if ("suffix" in self) and ("attach" in self):
            self._write_file(self.attach_path(self["suffix"]),
                             base64.decodestring(self["attach"]))
        elif "attach" in self:
            self._write_file(self.attach_path("txt"),
                             base64.decodestring(self["attach"]))

    def body_string(self):
        """Remove attach field."""
        buf = [self["stamp"], self["id"]]
        for k in self:
            if k in ("stamp", "id"):
                pass
            elif k == "attach":
                buf.append("attach:1")
            elif k == "sign":
                pass
            elif k == "pubkey":
                if self.check_sign():
                    short_key = apollo.cut_key(self["pubkey"])
                    buf.append("pubkey:" + short_key)
            else:
                buf.append(k + ":" + self[k])
        return "<>".join(buf)

    def check_sign(self):
        """Check sigunature."""
        for k in ("pubkey", "sign", "target"):
            if k not in self:
                return False
        target = ""
        for t in self["target"].split(","):
            try:
                target += "<>" + t + ":" + self[t]
            except KeyError:
                return False
            except ValueError:
                return False
        target = target[2:]         # remove ^<>

        md = md5.new(target).hexdigest()
        if apollo.verify(md, self["sign"], self["pubkey"]):
            return True
        else:
            return False

    def exists(self):
        return os.path.isfile(self.path)

    def removed(self):
        return os.path.isfile(self.rm_path)

    def size(self):
        try:
            return os.path.getsize(self.path)
        except OSError, err:
            sys.stderr.write('OSError: %s\n' % err)
            return 0

    def remove(self):
        """Remove record.

        The record is removed from cache,
        but do not removed from file-system.
        """
        try:
            shutil.move(self.path, self.rm_path)
            for path in (self.body_path, self.attach_path()):
                if path and os.path.isfile(path):
                    os.remove(path)
            return True
        except (IOError, OSError), err:
            sys.stderr.write('IOError/OSError: %s\n' % err)
            return False

# End of Record


class RecordGetter:
    '''Iterator to get records with heads.
    '''
    def __init__(self, datfile, node, head):
        self.datfile = datfile
        self.node = node
        self.head = head

    def __iter__(self):
        for h in self.head:
            r = Record(datfile = self.datfile,
                       idstr = h.strip().replace('<>', '_'))
            if not (r.exists() or r.removed()):
                try:
                    res = self.node.talk('/get/%s/%d/%s' %
                                    (self.datfile, r.stamp, r.id))
                    first = re.sub(r'[\r\n]*$', '', iter(res).next())
                    yield first
                except StopIteration, err:
                    sys.stderr.write('get %s: %s\n' % (self, err))

# End of RecordGetter


class Cache(dict):

    """Cache of BBS.

    Plain text (encode: UTF-8).
    One record par one line.
    """

    datfile = ""
    datpath = config.cache_dir
    stamp = 0       # when the cache is modified
    size = 0        # size of cache file
    type = ""       # "thread"

    def __init__(self, datfile):
        dict.__init__(self)
        self.datfile = datfile
        self.datpath += "/" + datfile
        self.removed = {}

        self.stamp = self._load_status("stamp")
        self.size = self._load_status("size")
        self.node = ListFile("%s/node.txt" % self.datpath)

        for type in config.types:
            if self.datfile.startswith(type):
                self.type = type
                break

        if self.exists():
            try:
                for k in listdir(self.datpath + "/record"):
                    self[k] = Record(datfile=self.datfile, idstr=k)
            except OSError:
                sys.stderr.write("%s/record: OSError\n" % self.datpath)

        self.save_record = config.save_record.get(self.type, 0)
        self.get_range = config.get_range.get(self.type, 0)
        self.sync_range = config.sync_range.get(self.type, 0)
        self.save_removed = config.save_removed.get(self.type, 0)

        if self.sync_range == 0:
            self.save_removed = 0
        elif self.save_removed == 0:
            pass
        elif self.save_removed <= self.sync_range:
            self.save_removed = self.sync_range + 1

    def __str__(self):
        return self.datfile

    def keys(self):
        k = dict.keys(self)
        k.sort()
        return k

    def __iter__(self):
        for idstr in self.keys():
            yield self[idstr]

    def exists(self):
        return os.path.isdir(self.datpath)

    def has_record(self):
        removed = self.datpath + "/removed"
        return bool(self) or \
               (os.path.exists(removed) and
                bool(listdir(removed)))

    def _load_status(self, key):
        path = "%s/%s.stat" % (self.datpath, key)
        try:
            f = file(path)
            v = f.readline()
            f.close()
            return int(v.strip())
        except IOError:
            #sys.stderr.write(path + ": IOError\n")
            return 0
        except ValueError:
            sys.stderr.write(path + ": ValueError\n")
            return 0

    def _save_status(self, key, val):
        path = "%s/%s.stat" % (self.datpath, key)
        try:
            f = file(path, "wb")
            f.write(str(val)+"\n")
            f.close()
            return True
        except IOError:
            sys.stderr.write(path + ": IOError\n")
            return False

    def sync_status(self):
        self._save_status("stamp", self.stamp)
        self._save_status("size", self.size)

    def standby_directories(self):
        for d in ('', '/attach', '/body', '/record', '/removed'):
            if not os.path.isdir(self.datpath + d):
                try:
                    os.makedirs(self.datpath + d)
                except (IOError, OSError):
                    sys.stderr.write(self.datfile + ": IOError/OSError\n")

    def check_data(self, res, stamp=None, id=None, begin=None, end=None):
        '''Check a data and add it cache.'''
        flag_got = False
        flag_spam = False
        count = 0
        for i in res:
            count += 1
            r = Record(datfile=self.datfile)
            parse_ok = r.parse(i)
            if parse_ok and \
               ((stamp is None) or (r['stamp'] == str(stamp))) and \
               ((not id) or (r['id'] == id)) and \
               ((begin is None) or (begin <= r.stamp)) and \
               ((end is None) or (r.stamp <= end)) and \
               r.md5check():
                flag_got = True
                if (len(i) > config.record_limit*1024) or spam.check(i):
                    sys.stderr.write(
                        'Warning: %s/%s: too large or spam record.\n' %
                        (self.datfile, r.idstr))
                    self.add_data(r, False)
                    r.remove()
                    flag_spam = True
                else:
                    self.add_data(r)
            else:
                if stamp is not None:
                    str_stamp = '/%s' % stamp
                elif 'stamp' in r:
                    str_stamp = '/%s' % r['stamp']
                else:
                    str_stamp = ''
                sys.stderr.write("Warning: %s%s: broken record.\n" %
                                 (self.datfile, str_stamp))
            r.free()
        return count, flag_got, flag_spam

    def get_data(self, stamp=0, id="", node=None):
        """Get appointed data."""
        res = node.talk("/get/" + self.datfile + "/" + str(stamp) + "/" + id)
        count, flag_got, flag_spam = self.check_data(res, stamp=stamp, id=id)
        if count:
            self.sync_status()
        else:
            sys.stderr.write("Warning: %s/%s: records not found.\n" %
                             (self.datfile, stamp))
        return flag_got, flag_spam

    def get_with_range(self, node=None):
        """Get data in range."""
        now = int(time())
        if self.stamp > 0:
            begin = self.stamp
        else:
            begin = 0
        if self.sync_range > 0:
            begin2 = now - self.sync_range
        else:
            begin2 = 0
        if begin < 0:
            begin = 0
        elif begin2 < begin:
            begin = begin2

        if (begin <= 0) and (len(self) <= 0):
            if self.get_range > 0:
                begin = now - self.get_range
                if begin < 0:
                    begin = 0
            else:
                begin = 0
            res = node.talk('/get/%s/%d-' % (self.datfile, begin))
        else:
            head = node.talk('/head/%s/%d-' % (self.datfile, begin))
            res = RecordGetter(self.datfile, node, head)

        count = self.check_data(res, begin=begin, end=now)
        if count:
            self.sync_status()
        return bool(count)

    def add_data(self, rec, really=True):
        """Add new data cache."""
        self.standby_directories()
        rec.sync()
        if really:
            self[rec.idstr] = rec
            self.size += len(str(rec)) + 1
        if self.stamp < rec.stamp:
            self.stamp = rec.stamp

    def check_body(self):
        '''Remove body cache that is a field of removed record.'''
        try:
            dir = os.path.join(config.cache_dir, self.datfile, 'body')
            for idstr in listdir(dir):
                rec = Record(datfile=self.datfile, idstr=idstr)
                if not rec.exists():
                    try:
                        os.remove(os.path.join(dir, idstr))
                    except OSError, err:
                        sys.stderr.write("%s/%s: OSError: %s\n" %
                                         (dir, idstr, err))
        except (IOError, OSError), err:
            sys.stderr.write('IOError/OSError: %s\n' % err)

    def check_attach(self):
        """Remove attach cache that is a field of removed record."""
        try:
            dir = os.path.join(config.cache_dir, self.datfile, 'attach')
            for f in listdir(dir):
                idstr = f
                i = f.find(".")
                if i >= 0:
                    idstr = f[:i]
                rec = Record(datfile=self.datfile, idstr=idstr)
                if not rec.exists():
                    try:
                        os.remove(dir + "/" + f)
                    except OSError, err:
                        sys.stderr.write('OSError: %s\n' % err)
        except (IOError, OSError), err:
            sys.stderr.write('IOError/OSError: %s\n' % err)

    def remove(self):
        """Remove cache (a.k.a DATFILE).

        It is removed from disk.
        """
        try:
            shutil.rmtree(self.datpath)
            return True
        except IOError, strerror:
            sys.stderr.write("IOError: " + strerror + "\n")
            return False

    def remove_records(self, now, limit):
        '''Remove records which are older than limit.
        '''
        ids = self.keys()
        while ids:
            r = ids.pop()
            rec = self[r]
            if not (("remove_stamp" in rec) or \
                    ("remove_id" in rec)):
                break
            elif len(self) <= 0:
                break
        once = {}
        for r in ids:
            rec = self[r]
            if ((limit > 0) and (rec.stamp + limit < now)) or \
               (rec.id in once):
                rec.remove()
            else:
                once[rec.id] = True

    def search(self, searchlist=None, myself=None):
        """Search node from network and get records."""
        self.standby_directories()
        if searchlist is None:
            searchlist = SearchList()
        if not myself:
            nodelist = NodeList()
            myself = nodelist.myself()
        node = searchlist.search(self, myself=myself)
        if node is not None:
            nodelist = NodeList()
            if node not in nodelist:
                nodelist.append(node)
                nodelist.sync()
            self.get_with_range(node)
            if node not in self.node:
                while len(self.node) >= config.share_nodes:
                    n = self.node.random()
                    self.node.remove(n)
                self.node.append(node)
                self.node.sync()
            return True
        else:
            self.sync_status()
            return False

# End of Cache


class CacheList(list):

    """All cache."""

    def __init__(self):
        list.__init__(self)
        for i in listdir(config.cache_dir):
            c = Cache(i)
            self.append(c)

    def getall(self, timelimit=0):
        """Search nodes and update my cache."""
        random.shuffle(self)
        nodelist = NodeList()
        myself = nodelist.myself()
        searchlist = SearchList()
        for cache in self:
            if int(time()) > timelimit:
                sys.stderr.write("client timeout\n")
                return
            elif not cache.exists():
                pass
            else:
                cache.search(searchlist=searchlist, myself=myself)
                cache.size = 0
                for rec in cache:
                    if not rec.exists():
                        continue
                    load_ok = rec.load()
                    if load_ok:
                        if cache.stamp < rec.stamp:
                            cache.stamp = rec.stamp
                        cache.size += len(str(rec))
                        rec.sync(True)
                        rec.free()
                    else:
                        rec.remove()
                cache.check_body()
                cache.check_attach()
                cache.sync_status()

    def search(self, query):
        result = []
        for cache in self:
            for rec in cache:
                try:
                    rec.load_body()
                    if query.search(unicode(str(rec), 'utf-8')):
                        result.append(cache)
                        rec.free()
                        break
                except (IOError, OSError, UnicodeDecodeError):
                    pass
                rec.free()
        return result

    def clean_records(self):
        """Clean old records."""
        now = int(time())
        for cache in self:
            cache.remove_records(now, cache.save_record)

    def remove_removed(self):
        """Remove removed record from disk."""
        now = int(time())
        for cache in self:
            for r in listdir(cache.datpath + "/removed"):
                rec = Record(datfile=cache.datfile, idstr=r)
                if (cache.save_removed > 0) and \
                   (rec.stamp + cache.save_removed < now) and \
                   (rec.stamp < cache.stamp):
                    try:
                        os.remove(cache.datpath + "/removed/" + r)
                    except OSError, (errno, errorstr):
                        sys.stderr.write("OSError: %s: %s\n" %
                                         (cache.datpath + "/removed/" + r,
                                          errorstr))

# End of CacheList


cached_update = {}

class UpdateList(list):

    """Save update information.

    Get update if the system does not have the update.
    """

    def __init__(self,
                 update_file = config.update,
                 update_range = config.update_range):
        list.__init__(self)
        self.update_file = update_file
        self.update_range = update_range
        if update_file in cached_update:
            for r in cached_update[update_file]:
                self.append(r)
        else:
            try:
                f = file(self.update_file)
                for line in f:
                    r = self.make_record(line)
                    if r is not None:
                        self.append(r)
                f.close()
            except IOError:
                sys.stderr.write(self.update_file + ": IOError\n")

    def make_record(self, line):
        buf = re.sub(r'[\r\n]*$', '', line).split("<>")
        if len(buf) > 2:
            idstr = buf[0] + "_" + buf[1]
            r = Record(datfile=buf[2], idstr=idstr)
            return r
        else:
            return None

    def sync(self, target=None):
        """Save update list."""
        if target is None:
            target = self
        f = file(self.update_file, "wb")
        now = int(time())
        once = {}
        buf = []
        for r in target:
            if (self.update_range == 0) or \
               (now - self.update_range < r.stamp):
                line = "<>".join((str(r.stamp), r.id, r.datfile)) + "\n"
                if line not in once:
                    f.write(line)
                    buf.append(r)
                    once[line] = True
        cached_update[self.update_file] = buf
        f.close()

# End of UpdateList


class RecentList(UpdateList):
    """Recent updated files list.

    Provide file names for node manager.
    """

    def __init__(self):
        UpdateList.__init__(self, config.recent, config.recent_range)

    def getall(self):
        """Search nodes and update my recent list."""
        searchlist = SearchList()
        if config.recent_range:
            begin = int(time()) - config.recent_range
        else:
            begin = 0
        count = 0
        for node in searchlist:
            res = node.talk("/recent/%d-" % begin)
            for line in res:
                if not re.search(r'^\d+<>[0-9a-f]{32}<>[_0-9A-Za-z]+$', line):
                    continue
                r = self.make_record(line)
                if r is not None:
                    self.append(r)
            count += 1
            if count > config.search_depth:
                break
        buf = self.uniq()
        self.sync(buf)

    def uniq(self):
        '''Save one line par one file.'''
        date = {}
        for r in self:
            if (r.datfile not in date) or (date[r.datfile] < r):
                date[r.datfile] = r
        buf = []
        for datfile in date:
            buf.append(date[datfile])
        return buf

# End of RecentList
